import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

class DjinnService {
  Map<String, dynamic> state;

  List<UpdateCallback> _listeners = [];
  List<State> _widgets = [];

  void init() {}

  void patch(Map<String, dynamic> newState) {
    state = state ?? {};
    Map<String, StateUpdate> update = {};

    newState.keys.forEach((String key) {
      update[key] = StateUpdate(newState[key], state[key]);
      state[key] = newState[key];
    });

    _callListeners(update);
  }

  void listen<T extends State>(T widget) {
    _widgets.add(widget);
  }

  void dispose<T extends State>(T widget) {
    _widgets.remove(widget);
  }

  Function subscribe(UpdateCallback fn) {
    _listeners.add(fn);
    return () => _removeListener(fn);
  }

  _removeListener(UpdateCallback fn) {
    _listeners.remove(fn);
  }

  void _callListeners(Map<String, StateUpdate> update) {
    _listeners.forEach((UpdateCallback it) => it(update));
    // ignore: INVALID_USE_OF_PROTECTED_MEMBER
    _widgets.forEach((State it) => it.setState(() {}));
  }
}

typedef UpdateCallback(Map<String, StateUpdate> update);
typedef void SetStateFn(VoidCallback fn);

class StateUpdate {
  dynamic current;
  dynamic previous;

  StateUpdate(this.current, this.previous);
}
